import pandas as pd
from scipy.optimize import minimize


class ModelCalibration:
    """
    Calibrate option pricing models according to prices of listed options

    :param options: a collection of :class:`~option.VanillaOption`
    """

    DEFAULT_PARAMETER_LOWER_BOUND = 1e-4

    def __init__(self, options):
        self.options = options

    def calibrate_model(self, model, method=None, options=None, bounds='default'):
        """
        Tune model parameters and returns a tuned model. The algorithm tries to minimize the squared difference
        between the prices of listed options and the prices predicted by the model, by tuning model parameters.
        The numerical optimization is performed by :func:`~scipy.optimize.minimize` in the scipy package.

        :param model: the model to calibrate
        :param method: see :func:`~scipy.optimize.minimize`
        :param options: see :func:`~scipy.optimize.minimize`
        :param bounds: the bounds to apply to parameters. If none is specified, then the
                       :attr:`~DEFAULT_PARAMETER_LOWER_BOUND` is applied for all the parameters.
                       Otherwise, a list of tuples (lower_bound, upper_bound) for each parameter shall be specified.
        :return: a tuple (res, model), where res is the result of :func:`~scipy.optimize.minimize`,
                 while res a calibrated model
        """
        if bounds == 'default':
            bounds = ((self.DEFAULT_PARAMETER_LOWER_BOUND, None),) * len(model.parameters)
        loss = self._get_loss_function(model)
        res = minimize(loss, model.parameters, bounds=bounds, method=method, options=options)
        model.parameters = res.x
        return res, model

    def _get_loss_function(self, model):
        def _loss_function(parameters):
            model.parameters = parameters
            predicted_prices = pd.Series([model.price_option_black(o) for o in self.options])
            real_prices = pd.Series([o.price for o in self.options])
            return sum((predicted_prices - real_prices) ** 2)

        return _loss_function



import abc

import numpy as np
from scipy import linalg as la

from vanilla_option_pricing.option_pricing import OptionPricingModel


class PossiblePricingModel(abc.ABC):
    """
    A model which can be used to price options, because it exposes the methods required by :class:`~option_pricing.OptionPricingModel`.
    """
    def as_option_pricing_model(self):
        """
        Converts the model to as option pricing model, thus providing pricing methods

        :return: an :class:`~option_pricing.OptionPricingModel` based on this model
        """
        return OptionPricingModel(self)


class LogMeanRevertingToGeneralisedWienerProcess(PossiblePricingModel):
    """
    The Log Mean-Reverting To Generalised Wiener Process model. It is a two-factor, mean reverting model, where the
    long-term behaviour is given by a Geometric Brownian motion, while the short-term mean-reverting tendency
    is modelled by an Ornstein-Uhlenbeck process.

    :param p_0: the initial variance, that is the variance of the state at time t=0. Must be a 2x2 numpy matrix
    :param l: the strength of mean-reversion
    :param s_x: volatility of the long-term process
    :param s_y: volatility of the short-term process
    """

    name = 'Log Mean-Reverting To Generalised Wiener Process'

    def __init__(self, p_0: np.matrix, l: float, s_x: float, s_y: float):
        self.p_0 = p_0
        self.l = l
        self.s_x = s_x
        self.s_y = s_y

    @property
    def parameters(self):
        """
        Model parameters, as a list of real numbers, in the order [l, s_x, s_y].
        """
        return [self.l, self.s_x, self.s_y]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s_x = value[1]
        self.s_y = value[2]

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        first_term = (self.p_0[0, 0] - 2 * self.p_0[1, 0] + self.p_0[1, 1] - (self.s_x ** 2 + self.s_y ** 2) / (
                2 * self.l)) * np.exp(-2 * self.l * t)
        second_term = 2 * (self.p_0[1, 0] - self.p_0[1, 1] + self.s_y ** 2 / self.l) * np.exp(-self.l * t)
        third_term = (self.s_y ** 2) * t + (self.s_x ** 2 - 3 * (self.s_y ** 2)) / (2 * self.l) + self.p_0[1, 1]
        return first_term + second_term + third_term


class OrnsteinUhlenbeck(PossiblePricingModel):
    """
    The single-factor, mean-reverting Ornstein-Uhlenbeck process.

    :param p_0: the initial variance, that is the variance of the state at time t=0
    :param l: the strength of the mean-reversion
    :param s: the volatility
    """
    name = 'Ornstein-Uhlenbeck'

    def __init__(self, p_0: float, l: float, s: float):
        self.p_0 = p_0
        self.l = l
        self.s = s

    @property
    def parameters(self):
        """
        Model parameters, as a list of real numbers, in the order [l, s].
        """
        return [self.l, self.s]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s = value[1]

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        return self.p_0 * np.exp(-2 * self.l * t) + self.s ** 2 / (2 * self.l) * (1 - np.exp(-2 * self.l * t))


class BlackScholes(PossiblePricingModel):
    """
    The famous Black-Sholes model, basically a Geometric Brownian Motion

    :param s: the volatility
    """
    name = 'Black-Sholes'

    def __init__(self, s: float):
        self.s = s

    @property
    def parameters(self):
        """
        Model parameters, as a list of real numbers, in the order [s].
        """
        return [self.s]

    @parameters.setter
    def parameters(self, value):
        self.s = value[0]

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        return self.s ** 2 * t


class NumericalLogMeanRevertingToGeneralisedWienerProcess(PossiblePricingModel):
    """
    This model relies on the same stochastic process as :class:`~models.LogMeanRevertingToGeneralisedWienerProcess`,
    but uses numerical procedures based on matrix exponential instead of closed formulas to compute the variance.
    As this approach is considerably slower, it is strongly suggested to use
    :class:`~models.LogMeanRevertingToGeneralisedWienerProcess` instead, using this class only for benchmarking

    :param p_0: the initial variance, that is the variance of the state at time t=0. Must be a 2x2 numpy matrix
    :param l: the strength of mean-reversion
    :param s_x: volatility of the long-term process
    :param s_y: volatility of the short-term process
    """

    name = 'Numerical Log Mean-Reverting To Generalised Wiener Process'

    def __init__(self, p_0: np.matrix, l: float, s_x: float, s_y: float):
        self.p_0 = p_0
        self.l = l
        self.s_x = s_x
        self.s_y = s_y
        self.numerical_model = NumericalModel(self.__get_A_matrix(), self.__get_B_matrix(), self.p_0)

    @property
    def parameters(self):
        """
        Model parameters, as a list of real numbers, in the order [l, s_x, s_y].
        """
        return [self.l, self.s_x, self.s_y]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s_x = value[1]
        self.s_y = value[2]
        self.numerical_model.parameters = [self.__get_A_matrix(), self.__get_B_matrix()]

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        return self.numerical_model.variance(t)

    def __get_A_matrix(self):
        return np.matrix([[-self.l, self.l], [0, 0]])

    def __get_B_matrix(self):
        return np.matrix([[self.s_x, 0], [0, self.s_y]])


class NumericalModel:
    """
    A general-purpose linear stochastic system. All the parameters must be matrices of suitable dimension

    :param A: the dynamic matrix A of the system
    :param B: the input matrix B of the system
    :param p_0: the initial variance, that is the variance of the state at time t=0

    """

    def __init__(self, A: np.matrix, B: np.matrix, p_0: np.matrix):
        self.A = A
        self.B = B
        self.p_0 = p_0

    @property
    def parameters(self):
        """
        Model parameters, as a list of matrices, in the order [A, B].
        """
        return [self.A, self.B]

    @parameters.setter
    def parameters(self, value):
        self.A = value[0]
        self.B = value[0]

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        dim = self.A.shape[0]
        F = la.expm(
            np.bmat([
                [self.A, self.B * np.transpose(self.B)],
                [np.zeros_like(self.A), -np.transpose(self.A)]
            ]) * t)
        P = (F[0:dim, 0:dim] * self.p_0 + F[0:dim, dim:2 * dim]) * la.inv(F[dim:2 * dim, dim:2 * dim])
        return P[0, 0]



import numpy as np
from py_vollib.black import undiscounted_black
from py_vollib.black_scholes_merton import black_scholes_merton


class OptionPricingModel:
    """
    A model which can be used to price European vanilla options.

    :param model: the stochastic model of the underlying
    """

    def __init__(self, model):
        self.model = model

    @property
    def parameters(self):
        """
        The model parameters, returned as a list of values
        """
        return self.model.parameters

    @parameters.setter
    def parameters(self, value):
        self.model.parameters = value

    def variance(self, t):
        """
        The variance of the model output at a certain time instant

        :param t: the time when the variance is evaluated
        :return: the variance at time t
        """
        return self.model.variance(t)

    def standard_deviation(self, t):
        """
        The standard deviation of the model output at a certain time instant,
        that is the squared root of the :func:`~option_pricing.OptionPricingModel.variance`
        at the same time instant

        :param t: the time when the standard deviation is evaluated
        :return: the standard deviation at time t
        """
        return np.sqrt(self.variance(t))

    def volatility(self, t):
        """
        The volatility of the model output at a certain time instant,
        that is the :func:`~option_pricing.OptionPricingModel.standard_deviation`
        divided by the squared root of the time

        :param t: the time when the volatility is evaluated
        :return: the volatility at time t
        """
        return self.standard_deviation(t) / np.sqrt(t)

    def price_black_scholes_merton(self, option_type, spot, strike, years_to_maturity, risk_free_rate, dividend=0):
        """
        Finds the no-arbitrage price of a European Vanilla option. Price is computed using the Black-Scholes-Merton
        formulae, but the variance of the underlying is extracted from this model.

        :param option_type: the type of the option (c for call, p for put)
        :param spot: the spot price of the underlying
        :param strike: the option strike price
        :param years_to_maturity: the years remaining before maturity - as a decimal number
        :param risk_free_rate: the risk-free interest rate
        :param dividend: the dividend paid by the underlying - as a decimal number
        :return: the no-arbitrage price of the option
        """
        volatility = self.volatility(years_to_maturity)
        price = black_scholes_merton(option_type, spot, strike, years_to_maturity, risk_free_rate, volatility, dividend)
        return price

    def price_option_black_scholes_merton(self, option, risk_free_rate: float):
        """
        Same as :func:`~option_pricing.OptionPricingModel.price_black_scholes_merton`, but the details
        of the vanilla option are provided by a :class:`~option.VanillaOption` object

        :param option: a :class:`~option.VanillaOption`
        :param risk_free_rate: the risk-free interest rate
        :return: the no-arbitrage price of the option
        """
        return self.price_black_scholes_merton(
            option.option_type,
            option.spot,
            option.strike,
            option.years_to_maturity,
            risk_free_rate,
            option.dividend
        )

    def price_black(self, option_type, spot, strike, years_to_maturity):
        """
        Finds the no-arbitrage price of a European Vanilla option. Price is computed using the Black
        formulae, but the variance of the underlying is extracted from this model.

        :param option_type: the type of the option (c for call, p for put)
        :param spot: the spot price of the underlying
        :param strike: the option strike price
        :param years_to_maturity: the years remaining before maturity - as a decimal number
        :return: the no-arbitrage price of the option
        """
        volatility = self.volatility(years_to_maturity)
        price = undiscounted_black(spot, strike, volatility, years_to_maturity, option_type)
        return price

    def price_option_black(self, option):
        """
        Same as :func:`~option_pricing.OptionPricingModel.price_black`, but the details
        of the vanilla option are provided by a :class:`~option.VanillaOption` object

        :param option: a :class:`~option.VanillaOption`
        :return: the no-arbitrage price of the option
        """
        return self.price_black(option.option_type, option.spot, option.strike, option.years_to_maturity)
